/*******************************************************************************
*                          						       *
*                                  magnum v0.7 	                               *
*                                                                              *
*                            (c) 1996  Brian Stempel                           *
*                          						       *
*  Please report questions, comments, bugs to stempel@oneimage.com.  This      *
*  software is distributed under the GNU General Public License, and hence     *
*  comes with absolutely *NO* warranty.                                        *
*                                                                              *
*******************************************************************************/

#ifdef AIX
#define EARGS "/usr/bin/ps", "ps", "-eo", "user,pid,args"
#define SARGS "/usr/bin/ps", "ps", "-eo", "user,pid,command"
#define MARGS "/bin/mail"
#endif
#ifdef Linux
#define EARGS "/bin/ps", "ps", "-aux"
#define SARGS "/bin/ps", "ps", "-auxc"
#define MARGS "/bin/mail"
#endif
#ifdef SunOS
#define EARGS "/bin/ps", "ps", "-aux"
#define SARGS "/bin/ps", "ps", "-auxc"
#define MARGS "/usr/ucb/mail"
#endif

#include <stdio.h>
#include <pwd.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

/* If you still want some kind of .plan, change below */
#define REALPLAN   "No plan.\n"

/* This affects ps output, if you anticipate ps outputting more */
/* than 1024 lines of 128 characters each, change below         */
#define LINELENGTH 128
#define MAXLINES 1024

#define FRAG "fingered you by the name"

int match(char *is_user);

int is_command(char *cmd, char *is_cmd);

void deliver(char *subj, char *body);

void breakhandler();

char *user,
     plandir[32],
     planpath[256];

struct passwd *db_record;

void main() {

	FILE *fp;

	time_t UTC_time;
 
	char *tok,
	     subj[256],
	     body[256],
	     *owner,
	     *finger_arg,
	     hostnam[256],
	     *fingerer_name,
	     *formatted_time,
	     line_buff[MAXLINES][LINELENGTH];

	pid_t pid;
	
	int i,
	    id,
	    stream,
	    fd_rw[2],
	    curr_line;
            
	/* Get user/system info */
	gethostname(hostnam, sizeof(hostnam));
	db_record = getpwuid(getuid());
	user = db_record->pw_name;

	/* Determine run status */
	if (pipe(fd_rw) < 0) {

		perror("magnum: runstatus: pipe");
		exit(1);

	}

	if ((id = fork()) < 0) {

		perror("magnum: runstatus: fork");
		exit(1);

	}
	
	if (id == 0) {

		dup2(fd_rw[1], 1);
		close(fd_rw[1]);
		close(fd_rw[0]);
		execl(SARGS, (char *) 0);
		perror("magnum: exec");
		exit(1);

	}

	close(fd_rw[1]);

	fp = fdopen(fd_rw[0], "r");

	/* Get lines, one at a time, from ps output */
	curr_line = 0;
	while (fgets(line_buff[curr_line], LINELENGTH, fp) != NULL)
		++curr_line;
	
	/* Close stream and see if we're already running */
	fclose(fp);
	close(fd_rw[0]);
	while (wait((int *) 0) != id);
	for (i=0;i<=curr_line - 1;++i) {

		owner = strtok(line_buff[i]," ");
		pid = (pid_t) atoi(strtok(NULL, " "));
		if (!(strcmp(owner, user))) {

			while ((tok = strtok(NULL, " ")) != NULL) {

				if (is_command("magnum", tok)) {

					if (getpid() != pid)
						kill(pid, SIGTERM);

				}

			}

		}

	}
	
	/* Remove dir, if previous runs didn't clean up */
	sprintf(plandir, "/tmp/.%s", user);
	sprintf(planpath, "/tmp/.%s/.plan", user);
	remove(planpath);
	rmdir(plandir);

	/* Catch certain signals, so cleanup is done upon exit */
	signal(SIGQUIT, breakhandler);
	signal(SIGINT, breakhandler);
	signal(SIGTERM, breakhandler);
	signal(SIGSTOP, breakhandler);

	/* Main loop */
	while (1) {

		/* Make .plan dir */
		mkdir(plandir, S_IRUSR|S_IWUSR|S_IXUSR|S_IXGRP|S_IXOTH);

		/* Setup named pipe */
		mkfifo(planpath, S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
        	
		/* Block on write until someone fingers */
		if ((stream = open(planpath, O_WRONLY)) == -1) {

			perror("magnum: fopen");
			exit(1);

		}

		write(stream, REALPLAN, sizeof(REALPLAN)-1);

		if (pipe(fd_rw) < 0) {

			perror("magnum: pipe");
			exit(1);

		}

		if ((id = fork()) < 0) {

			perror("magnum: fork");
			exit(1);

		}

		if (id == 0) {

			dup2(fd_rw[1], 1);
			close(fd_rw[1]);
			close(fd_rw[0]);
			execl(EARGS, (char *) 0);
			exit(0);

		}

		close(fd_rw[1]);

		fp = fdopen(fd_rw[0], "r");

		/* Fill buffer with ps output */
		curr_line = 0;
 		while (fgets(line_buff[curr_line], LINELENGTH, fp) != NULL)
			++curr_line;
		
		/* Close stream to return user ASAP to prompt, then process */
		close(stream);
		remove(planpath);
		fclose(fp);
		close(fd_rw[0]);
		while (wait((int *) 0) != id);

		/* Parse each line -- a little too nested, but short */
		for (i=0;i<=curr_line - 1;++i) {

			fingerer_name = strtok(line_buff[i]," ");
			while ((tok = strtok(NULL, " ")) != NULL) {

				if (is_command("finger", tok)) {

					if ((finger_arg = strtok(NULL,"'\n',' '"
					)) != NULL) {

						/* Mail user about finger */
						if ((match(finger_arg) && strcmp
						(fingerer_name, "nobody"))) {

							UTC_time = time(
							(time_t *) NULL);
							formatted_time = asctime
							(localtime(&UTC_time));
							sprintf(body,
							"User %s %s %s on %s",
							fingerer_name, FRAG,
							finger_arg,
							formatted_time);
							sprintf(subj,
							"magnum (%s): %s",
							hostnam, fingerer_name);
							deliver(subj, body);

						}

					}

				}

			}

		}
		/* Look familiar? */

	}

}

void deliver(char *subj, char *body) {

	int id,
	    mail_rw[2];
	
	FILE *fp;

	if (pipe(mail_rw) < 0) {

		perror("magnum: mail: pipe");
		exit(1);

	}

	if ((id = fork()) < 0) {

	perror("magnum: mail: fork");
		exit(1);

	}

	if (id == 0) {

		dup2(mail_rw[0], 0);
		close(mail_rw[0]);
		close(mail_rw[1]);
		execl(MARGS, "mail", "-s", subj, user, (char *) 0);

	}

	close(mail_rw[0]);
	fp = fdopen(mail_rw[1], "w");
	fprintf(fp, "%s", body);
	fclose(fp);
	close(mail_rw[1]);
	while (wait((int *) 0) != id);
	
}

int match(char *is_user) {

	size_t tok_len,
	       user_len;

	char *token,
	     *full_name,
	     lowercase_user[64],
	     lowercase_login[16];

	int i;

	/* 'Lowercase' the fingerer's 'finger' arg */
	strcpy(lowercase_user, is_user);
	user_len = strlen(lowercase_user);
	for (i=0;i<=user_len-1;++i)
		lowercase_user[i] = tolower(lowercase_user[i]);
	lowercase_user[i+1] = 0x0;

	/* 'Lowercase' the user's login (usually not necessary) */
	strcpy(lowercase_login, user);
	user_len = strlen(lowercase_login);
	for (i=0;i<=user_len-1;++i)
		lowercase_login[i] = tolower(lowercase_login[i]);
	lowercase_login[i+1] = 0x0;

	/* Match actual login, whether upper or lower case */
	if (!strcmp(lowercase_login, lowercase_user))
		return 1;

	/* Match any token of user's real name, whether upper or lower case */
	if ((full_name = strtok(db_record->pw_gecos, ",")) != NULL) {

		token = strtok(full_name, " ");
		do {

			tok_len = strlen(token);
			for (i=0;i<=tok_len-1;++i)
				token[i] = tolower(token[i]);
			token[i+1] = 0x0;
			if (!strcmp(token, lowercase_user))
				return 1;

		} while ((token = strtok(NULL, " ")) != NULL);

	}

	return 0;

}

int is_command(char *cmd, char *is_cmd) {

	int i = 0;

	char tmp_cmd[64];

	while((is_cmd[i] != '\n') && (is_cmd[i] != ' ')) {
		tmp_cmd[i] = is_cmd[i];
		++i;
	}
	tmp_cmd[i] = 0x0;

	if (!(strcmp(cmd, &tmp_cmd[strlen(tmp_cmd) - strlen(cmd)])))
		return 1;

	return 0;
        
}

void breakhandler(void) {

	remove(planpath);
	rmdir(plandir);
	exit(1);

}
